/* 
 * Sound Monitor applet for the GNOME panel
 * (C) 2002 John Ellis
 *
 * Author: John Ellis
 *
 * This software is released under the GNU General Public License (GNU GPL).
 * Please read the included file COPYING for more information.
 * This software comes with no warranty of any kind, use at your own risk!
 */

/*
 * Portions of code in this file originally based on vu-meter by Gregory McLean
 * almost completely different by now.
 */

#include "sound-monitor.h"
#include "esdcalls.h"


#define BUFFER_SIZE(x) (x * NSAMP * sizeof(short))
#define BUFFER_OFFSET(x) (x * NSAMP)

/* these are the times for the latency stuff */
#define SOUND_CHECK_NORMAL 60
#define SOUND_CHECK_FAST 23


static gint sound_check_set_speed(SoundData *sd, gint delay);
static gint esd_latency(gchar *host);


static gint open_sound (gchar *esd_host, gint use_input)
{
	gint sound = -1;

	if (esd_host && strlen(esd_host) == 0) esd_host = NULL;

	if (!use_input)
		{
		sound = esd_monitor_stream (ESD_BITS16|ESD_STEREO|ESD_STREAM|ESD_PLAY,
					    RATE, esd_host, "volume_meter_applet");
		}
	else
		{
		sound = esd_record_stream (ESD_BITS16|ESD_STEREO|ESD_STREAM|ESD_PLAY|ESD_RECORD,
					   RATE, esd_host, "volume_meter_applet");
		}

	if (sound < 0)
		{
		GtkWidget *box;
		box = gnome_error_dialog("Cannot connect to the sound daemon.\nThe sound daemon can be started by selecting 'Start ESD' from\n the right click menu, or by running 'esd' at a command prompt.");
		}

	return sound;
}

static void sound_compute_volume(SoundData *sd)
{
	short *active_buffer;
	unsigned short l, r;
	unsigned short val_l, val_r;
	int i;

	active_buffer = sd->buffer + BUFFER_OFFSET(sd->buffer_current);

	l = sd->left;
	r = sd->right;
	for (i = 1; i < NSAMP / 2;)
		{
		val_r = abs (active_buffer[i++]);
		val_l = abs (active_buffer[i++]);
		l = (val_l > l) ? val_l : l;
		r = (val_r > r) ? val_r : r;
		}
	l /= 327;
	r /= 327;

	sd->left = l;
	sd->right = r;
}

static void sound_read(gpointer data, gint source, GdkInputCondition condition)
{
	SoundData *sd = data;
	short *active_buffer;

	active_buffer = sd->buffer + BUFFER_OFFSET(sd->buffer_current);

	if (read(source, active_buffer, BUFFER_SIZE(1)) < 1)
		{
		return;
		printf("...nothing");
		}

	sd->buffer_current++;
	if (sd->buffer_current >= sd->buffer_count) sd->buffer_current = 0;

	sd->read_data = TRUE;
	sd->new_data = TRUE;

	sound_compute_volume(sd);

	return;
}

void sound_get_volume(SoundData *sd, gint *l, gint *r)
{
	if (!sd) return;

	*l = sd->left;
	*r = sd->right;

	/* this is reset after every call */
	sd->left = 0;
	sd->right = 0;
}

void sound_get_buffer(SoundData *sd, short **buffer, gint *length)
{
	if (!sd) return;

	*buffer = sd->buffer + BUFFER_OFFSET(sd->buffer_current);
	*length = NSAMP;
}

static gint sound_check_cb(gpointer data)
{
	SoundData *sd = data;

	if (sd->read_data)
		{
		sd->read_data = FALSE;
		sd->read_count = 0;
		return sound_check_set_speed(sd, SOUND_CHECK_NORMAL);;
		}

	if (sd->read_count >= sd->buffer_count)
		{
		sd->new_data = FALSE;
		return sound_check_set_speed(sd, SOUND_CHECK_NORMAL);
		}
	else
		{
		short *active_buffer;
		active_buffer = sd->buffer + BUFFER_OFFSET(sd->buffer_current);

		memset(active_buffer, 0, BUFFER_SIZE(1));

		sd->buffer_current++;
		if (sd->buffer_current >= sd->buffer_count) sd->buffer_current = 0;

		sd->new_data = TRUE;
		sd->read_count++;
		}

	return sound_check_set_speed(sd, SOUND_CHECK_FAST);
}

static gint sound_check_set_speed(SoundData *sd, gint delay)
{
	if (sd->check_id != -1)
		{
		if (delay > 20 && sd->check_speed == delay) return TRUE;
		gtk_timeout_remove(sd->check_id);
		sd->check_id = -1;
		}
	if (delay > 20)
		{
		sd->check_id = gtk_timeout_add(delay, sound_check_cb, sd);
		sd->check_speed = delay;
		}
	else
		{
		sd->check_speed = 0;
		}
	return FALSE;
}

static void sound_buffer_sync(SoundData *sd, guint count)
{
	g_free(sd->buffer);

	if (count < 1) count = 1;

	sd->buffer = g_malloc0(BUFFER_SIZE(count));
	sd->buffer_count = count;
	sd->buffer_current = 0;

	if (count < 2)
		{
		sound_check_set_speed(sd, -1);
		}
	else
		{
		sound_check_set_speed(sd, SOUND_CHECK_NORMAL);
		}

}

/* latency is 1000's of a second */
void sound_set_latency(SoundData *sd, gint latency)
{
	guint new_count;
	guint bufs_per_second;

	if (!sd) return;

	if (latency < 0)
		{
		latency = esd_latency(sd->host);
		}

	bufs_per_second = 1000 / (44100 / (NSAMP / 2));
	new_count = bufs_per_second * (guint)latency / 1000;
	if (new_count < 1) new_count = 1;

	sound_buffer_sync(sd, new_count);
}

/* returns TRUE if successful, FALSE if failed */
SoundData *sound_init(gchar *host, gint monitor_input)
{
	SoundData *sd;
	gint fd;

	fd = open_sound(host, monitor_input);

	if (fd < 1) return NULL;

	sd = g_new0(SoundData, 1);

	sd->buffer = NULL;
	sd->latency = 0;

	sd->host = g_strdup(host);

	sd->left = sd->right = 0;
	sd->new_data = FALSE;

	sd->read_data = FALSE;
	sd->read_count = FALSE;
	sd->check_speed = 0;
	sd->check_id = -1;

	sound_buffer_sync(sd, 1);

	sd->fd = fd;
	sd->input_cb_id = gdk_input_add (sd->fd, GDK_INPUT_READ, (GdkInputFunction)sound_read, sd);

	return sd;
}

void sound_free(SoundData *sd)
{
	if (!sd) return;

	sound_check_set_speed(sd, -1);

	if (sd->input_cb_id > -1)
		{
		gdk_input_remove(sd->input_cb_id);
		}
	if (sd->fd > -1)
		{
		esd_close(sd->fd);
		}

	g_free(sd->buffer);
	g_free(sd->host);

	g_free(sd);
}

/* esd controls */

gint esd_control(ControlType function, gchar *host)
{
	gint esd_fd;
	gint ret;

	switch (function)
		{
		case Control_Start:
			/* EEK! ESD should have a flag to start the daemon in
			 * the background, and return to cmd line when ready */
			ret = system("esd &");
			if (ret == -1 || ret == 127) return FALSE;
			sleep(3);
			return TRUE;
			break;
		case Control_Standby:
			esd_fd = esd_open_sound(host);
			if (esd_fd >= 0)
				{
				esd_standby(esd_fd);
				esd_close(esd_fd);
				return TRUE;
				}
			break;
		case Control_Resume:
			esd_fd = esd_open_sound(host);
			if (esd_fd >= 0)
				{
				esd_resume(esd_fd);
				esd_close(esd_fd);
				return TRUE;
				}
			break;
		}

	return FALSE;
}

StatusType esd_status(gchar *host)
{
	esd_standby_mode_t mode;
	gint esd_fd = 0;

	esd_fd = esd_open_sound(host);
	if (esd_fd>= 0)
		{
		mode = esd_get_standby_mode(esd_fd);
		esd_close(esd_fd);
		}
	else
		{
		/* hmm, well, this must be an error, right? */
		return Status_Error;
		}

	if (mode == ESM_ON_STANDBY) return Status_Standby;
	if (mode == ESM_ON_AUTOSTANDBY) return Status_AutoStandby;
	if (mode == ESM_RUNNING) return Status_Ready;

	return Status_Error;
}

static gint esd_latency(gchar *host)
{
	gint esd_fd = 0;
	gint ret = 0;

	esd_fd = esd_open_sound(host);
	if (esd_fd >= 0)
		{
		ret = esd_get_latency(esd_fd);
		esd_close(esd_fd);

		if (ret) ret = ret * 1000 / 44100;
		printf("latency result: %d\n", ret);
		}

	return ret;
}


