/*
 * (SLIK) SimpLIstic sKin functions
 * (C) 2002 John Ellis
 *
 * Author: John Ellis
 *
 * This software is released under the GNU General Public License (GNU GPL).
 * Please read the included file COPYING for more information.
 * This software comes with no warranty of any kind, use at your own risk!
 */

#include "ui2_includes.h"
#include "ui2_typedefs.h"

#include "ui2_meter.h"
#include "ui2_meter_edit.h"

#include "ui2_editor.h"
#include "ui2_widget.h"
#include "ui_fileops.h"
#include "ui_tabcomp.h"
#include "ui_utildlg.h"


typedef struct _MeterDetail MeterDetail;
struct _MeterDetail
{
	MeterData *meter;
	GtkWidget *segments_spin;
	GtkWidget *horizontal_toggle;
	GtkWidget *slide_toggle;
};

typedef struct _MeterListData MeterListData;
struct _MeterListData
{
	gchar *key;
	gchar *text_id;
	gchar *data;

	gchar *image;
	gint segments;
	gint horizontal;
	gint slide;
};

typedef struct _MeterPage MeterPage;
struct _MeterPage
{
	GtkWidget *key_entry;
	GtkWidget *text_id_entry;
	GtkWidget *data_entry;

	GtkWidget *image_entry;
	GtkWidget *segment_spin;
	GtkWidget *horizontal_toggle;
	GtkWidget *slide_toggle;

	GtkWidget *clist;
	GtkWidget *pixmap;

	EditData *ed;
};


static GdkPixbuf *meter_get_pixbuf(gpointer data)
{
	MeterData *meter = data;
	return meter->overlay;
}

static void meter_edit_write(FILE *f, WidgetData *wd, SkinData *skin, const gchar *dir)
{
	MeterData *meter = wd->widget;
	gchar *image;

	image = ui_edit_copy_unique_file(ui_widget_get_data(wd, "image"),
					 dir, meter->overlay, "meter_", wd->key);

	if (image) ui_edit_widget_set_path_key(wd, "image", dir, image);

	ui_edit_write_section(f, "meter", wd->key);

	ui_edit_write_key_char(f, "image", image);
	ui_edit_write_key_int(f, "x", meter->x);
	ui_edit_write_key_int(f, "y", meter->y);
	ui_edit_write_key_int(f, "segments", meter->segments);
	ui_edit_write_key_bool(f, "horizontal", meter->horizontal);
	ui_edit_write_key_bool(f, "slide", meter->slide);

	g_free(image);
}

static MeterListData *meter_edit_list_find(GList *list, const gchar *image)
{
	GList *work;
	work = list;
	while (work)
		{
		MeterListData *md = work->data;
		if (strcmp(image, md->image) == 0) return md;
		work = work->next;
		}
	return NULL;
}

static gpointer meter_edit_read(UIData *ui, WidgetData *wd, GList *list)
{
	MeterData *meter = wd->widget;
	MeterListData *md;
	const gchar *image;

	image = ui_widget_get_data(wd, "image");

	if (!image || meter_edit_list_find(list, image)) return NULL;

	md = g_new0(MeterListData, 1);
	md->image = g_strdup(image);
	md->segments = meter->segments;
	md->horizontal = meter->horizontal;
	md->slide = meter->slide;

	md->key = g_strdup(wd->key);
	md->data = g_strdup(ui_widget_get_data(wd, "data"));
	md->text_id = g_strdup(wd->text_id);

	return md;
}

static void meter_edit_props_segment_cb(GtkAdjustment *adjustment, gpointer data)
{
	MeterDetail *md = data;
	gint new_count;
	gint max_size;

	new_count = ui_edit_spin_get(md->segments_spin);

	if (md->meter->horizontal)
		{
		max_size = md->meter->width;
		}
	else
		{
		max_size = md->meter->height;
		}

	if (new_count > max_size)
		{
		new_count = max_size;
		ui_edit_spin_set_blocking(md->segments_spin, new_count, md);
		}

	md->meter->segments = new_count;
	md->meter->segment_size = max_size / new_count;
}

static void meter_edit_props_slide_cb(GtkWidget *widget, gpointer data)
{
	MeterDetail *md = data;

	md->meter->slide = ui_edit_toggle_get(md->slide_toggle);
}

static gpointer meter_edit_props(UIData *ui, WidgetData *wd, GtkWidget *vbox, gpointer detail)
{
	MeterData *meter = wd->widget;
	MeterDetail *md = detail;

	if (!md)
		{
		GtkObject *adj;

		md = g_new0(MeterDetail, 1);

		md->segments_spin = ui_edit_spin_new(vbox, _("Segments:"), 1, 2000, &adj);
		gtk_signal_connect(GTK_OBJECT(adj), "value_changed",
				   GTK_SIGNAL_FUNC(meter_edit_props_segment_cb), md);
		md->horizontal_toggle = ui_edit_toggle_new(vbox, _("Horizontal"));
		gtk_widget_set_sensitive(md->horizontal_toggle, FALSE);
		md->slide_toggle = ui_edit_toggle_new(vbox, _("Slide"));
		gtk_signal_connect(GTK_OBJECT(md->slide_toggle), "clicked",
				   GTK_SIGNAL_FUNC(meter_edit_props_slide_cb), md);
		}

	md->meter = meter;

	ui_edit_spin_set_blocking(md->segments_spin, meter->segments, md);
	ui_edit_toggle_set(md->horizontal_toggle, meter->horizontal);
	ui_edit_toggle_set(md->slide_toggle, meter->slide);

	return md;
}

static void meter_edit_free(gpointer data)
{
	MeterListData *md = data;
	g_free(md->key);
	g_free(md->text_id);
	g_free(md->data);
	g_free(md->image);
	g_free(md);
}

static void meter_edit_page_add_cb(GtkWidget *widget, gpointer data)
{
	MeterPage *pd = data;
	MeterData *meter;
	const gchar *key;
	const gchar *text_id;
	const gchar *image;

	key = ui_edit_entry_get(pd->key_entry);
	text_id = ui_edit_entry_get(pd->text_id_entry);
	image = ui_edit_entry_get(pd->image_entry);

	if (!key || !image || !isfile(image))
		{
		warning_dialog(_("Meter error"), _("Meter must contain a key and valid image."));
		return;
		}

	meter = meter_new_from_file(image, 0, 0,
				    ui_edit_spin_get(pd->segment_spin),
				    ui_edit_toggle_get(pd->horizontal_toggle),
				    ui_edit_toggle_get(pd->slide_toggle));
	if (meter)
		{
		WidgetData *wd;
		wd = meter_register(pd->ed->ui->skin, meter, key, text_id);
		ui_edit_widget_add_finish(pd->ed, wd, image, ui_edit_entry_get(pd->data_entry));
		}

	tab_completion_append_to_history(pd->image_entry, image);
}

static void meter_edit_page_sync(MeterPage *pd, MeterListData *md)
{
	if (!md) return;

	ui_edit_entry_set(pd->key_entry, md->key);
	ui_edit_entry_set(pd->text_id_entry, md->text_id);
	ui_edit_entry_set(pd->data_entry, md->data);

	ui_edit_entry_set(pd->image_entry, md->image);

	ui_edit_spin_set(pd->segment_spin, md->segments);
	ui_edit_toggle_set(pd->horizontal_toggle, md->horizontal);
	ui_edit_toggle_set(pd->slide_toggle, md->slide);
}

static void meter_edit_page_clist_cb(GtkWidget *clist, gint row, gint col, GdkEvent *event, gpointer data)
{
	MeterPage *pd = data;
	MeterListData *md;

	md = gtk_clist_get_row_data(GTK_CLIST(clist), row);
	meter_edit_page_sync(pd, md);
}

static void meter_edit_page_destroy_cb(GtkWidget *widget, gpointer data)
{
	MeterPage *pd = data;

	g_free(pd);
}

static GtkWidget *meter_edit_page_new(EditData *ed)
{
	GtkWidget *hbox;
	GtkWidget *vbox;
	GtkWidget *meter;
	MeterPage *pd;
	gchar *titles[] = { _("Image"), _("Key"), _("Segments"), NULL };

	pd = g_new0(MeterPage, 1);
	pd->ed = ed;

	hbox = gtk_hbox_new(FALSE, 5);
	gtk_container_set_border_width(GTK_CONTAINER(hbox), 5);
	gtk_object_set_data(GTK_OBJECT(hbox), "page", pd);
	gtk_signal_connect(GTK_OBJECT(hbox), "destroy",
			   GTK_SIGNAL_FUNC(meter_edit_page_destroy_cb), pd);

	vbox = gtk_vbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(hbox), vbox, FALSE, FALSE, 0);
	gtk_widget_show(vbox);

	pd->key_entry = ui_edit_key_entry_new(vbox, ed->ui, meter_type_id());
	pd->data_entry = ui_edit_entry_new(vbox, _("Data:"));
	pd->text_id_entry = ui_edit_entry_new(vbox, _("Text id:"));

	pd->segment_spin = ui_edit_spin_new(vbox, _("Segments:"), 1, 1000, NULL);
	pd->horizontal_toggle = ui_edit_toggle_new(vbox, _("Horizontal"));
	pd->slide_toggle = ui_edit_toggle_new(vbox, _("Slide"));

	pd->image_entry = ui_edit_path_entry_new(vbox, _("Image:"), "SLIK_meter_image");
	pd->pixmap = ui_edit_pixmap_new(vbox);
	ui_edit_path_entry_connect_pixmap(pd->image_entry, pd->pixmap);

	meter = gtk_button_new_with_label(_("Add"));
	gtk_box_pack_start(GTK_BOX(vbox), meter, FALSE, FALSE, 0);
	gtk_signal_connect(GTK_OBJECT(meter), "clicked",
			   GTK_SIGNAL_FUNC(meter_edit_page_add_cb), pd);
	gtk_widget_show(meter);

	pd->clist = ui_edit_clist_new(hbox, titles, 3);
	gtk_signal_connect(GTK_OBJECT(pd->clist), "select_row",
			   GTK_SIGNAL_FUNC(meter_edit_page_clist_cb), pd);

	gtk_widget_show(hbox);

	return hbox;
}

static void meter_edit_page_add(GtkWidget *widget, gpointer data)
{
	MeterListData *id = data;
	MeterPage *pd;
	gchar *buf[4];

	pd = gtk_object_get_data(GTK_OBJECT(widget), "page");

	buf[0] = "";
	buf[1] = id->key;
	buf[2] = ui_edit_int_to_text(id->segments);
	buf[3] = NULL;

	ui_edit_clist_append(pd->clist, buf, id->image, id);

	g_free(buf[2]);
}

void meter_type_init_edit(WidgetObjectData *od)
{
	od->func_get_pixbuf = meter_get_pixbuf;

	od->func_edit_write = meter_edit_write;

	od->func_edit_read = meter_edit_read;
	od->func_edit_free = meter_edit_free;

	od->func_edit_props = meter_edit_props;

	od->func_edit_page_new = meter_edit_page_new;
	od->func_edit_page_add = meter_edit_page_add;
}

