/*
 * (SLIK) SimpLIstic sKin functions
 * (C) 2002 John Ellis
 *
 * Author: John Ellis
 *
 * This software is released under the GNU General Public License (GNU GPL).
 * Please read the included file COPYING for more information.
 * This software comes with no warranty of any kind, use at your own risk!
 */

#include "ui2_includes.h"
#include "ui2_typedefs.h"
#include "ui2_scope.h"
#include "ui2_scope_edit.h"

#include "ui2_display.h"
#include "ui2_main.h"
#include "ui2_parse.h"
#include "ui2_skin.h"
#include "ui2_util.h"
#include "ui2_widget.h"
#include "ui_pixbuf_ops.h"

#include <math.h>

typedef struct _ScopeCallbackData ScopeCallbackData;
struct _ScopeCallbackData
{
	gint (*status_get_func)(ScopeData *scope, const gchar *key, gpointer data);
	gpointer status_get_data;
	void (*mode_get_func)(ScopeData *scope, const gchar *key, gpointer data);
	gpointer mode_get_data;
};

typedef struct _ScopeSetData ScopeSetData;
struct _ScopeSetData
{
	short *buffer;
	gint length;
	gint scale;
	ScopeRenderType mode;
};

static WidgetType type_id = -1;


/*
 *-----------------------------
 * sync function (public only for the editor)
 *-----------------------------
 */

void scope_sync_to_size(ScopeData *scope, gint w, gint h, gint horizontal)
{
	if (!scope || w < 4 || h < 4) return;

	scope->width = w;
	scope->height = h;
	scope->horizontal = horizontal;

	if (horizontal)
		{
		scope->points_count = scope->height;
		}
	else
		{
		scope->points_count = scope->width;
		}

	g_free(scope->points);
	scope->points = g_malloc(sizeof(gint) * scope->points_count);
	scope_widget_reset(scope);
	if (scope->pixbuf) gdk_pixbuf_unref(scope->pixbuf);
	scope->pixbuf = gdk_pixbuf_new(GDK_COLORSPACE_RGB, FALSE, 8,
				       scope->width, scope->height * 2);
}

/*
 *-----------------------------
 * new / free
 *-----------------------------
 */

ScopeData *scope_new(GdkPixbuf *pb, gint x, gint y, gint horizontal,
		     gint w, gint h, gint dynamic, gint stretch)
{
	ScopeData *scope;

	scope_type_init();

	util_size(&x);
	util_size(&y);
	util_size(&w);
	util_size(&h);

	if (!pb) return NULL;

	scope = g_new0(ScopeData, 1);

	scope->overlay = util_size_pixbuf(pb, TRUE);

	if (w < 4) w = gdk_pixbuf_get_width(scope->overlay);
	if (h < 4) h = gdk_pixbuf_get_height(scope->overlay) / 2;

	scope->width = w;
	scope->height = h;
	scope->x = x;
	scope->y = y;
	scope->horizontal = horizontal;
	scope->mode = SCOPE_RENDER_POINTS;
	scope->dynamic = dynamic;
	scope->stretch = stretch;

	scope->points = NULL;
	scope->pixbuf = NULL;

	scope_sync_to_size(scope, w, h, horizontal);

	return scope;
}

ScopeData *scope_new_from_data(gchar **data, gint x, gint y, gint horizontal,
			       gint w, gint h, gint dynamic, gint stretch)
{
	GdkPixbuf *pb;

	pb = gdk_pixbuf_new_from_xpm_data((const char **)data);

	return scope_new(pb, x, y, horizontal, w, h, dynamic, stretch);
}

ScopeData *scope_new_from_file(const gchar *file, gint x, gint y, gint horizontal,
		 	       gint w, gint h, gint dynamic, gint stretch)
{
	GdkPixbuf *pb;

	pb = gdk_pixbuf_new_from_file(file, NULL);

	return scope_new(pb, x, y, horizontal, w, h, dynamic, stretch);
}

void scope_free(ScopeData *scope)
{
	if (!scope) return;

	if (scope->pixbuf) gdk_pixbuf_unref(scope->pixbuf);
	if (scope->overlay) gdk_pixbuf_unref(scope->overlay);

	g_free(scope->points);
	g_free(scope);
}

static void scope_free_cb(gpointer data)
{
	scope_free((ScopeData *)data);
}

/*
 *-----------------------------
 * sound data parser
 *-----------------------------
 */

static gint scope_parse_buffer(ScopeData *scope, short *buffer, gint length)
{
	gint ret = FALSE;
	gint scale;
	gint step;
	gfloat fact;
	gint end;
	gint i, p;

	if (!buffer || length < scope->points_count) return scope_widget_reset(scope);

	if (length / 2 >= scope->scale * scope->points_count)
		{
		scale = scope->scale;
		}
	else
		{
		scale = length / 2 / scope->points_count;
		}

	step = scale * 2;

	if (scope->horizontal)
		{
		fact = (gfloat)65536 * scale / scope->width;
		}
	else
		{
		fact = (gfloat)65536 * scale / scope->height;
		}

	p = 0;
	end = scope->points_count * step;
	for (i = 0; i < end; i += step)
		{
		gint new_val = 0;
		gint k;

		for (k = 0; k < step; k += 2)
			{
			new_val += (buffer[i + k] + buffer[i + k + 1]) / 2;
			}
		new_val = (gint)((gfloat)new_val / fact);

		if (scope->points[p] != new_val)
			{
			scope->points[p] = new_val;
			ret = TRUE;
			}
		p++;
		}

	return ret;
}

/*
 *-----------------------------
 * draw
 *-----------------------------
 */

static void scope_draw_real(ScopeData *scope, GdkPixbuf *pb, UIData *ui)
{
	gint i;

	if (!scope) return;

	pixbuf_copy_area(scope->pixbuf, 0, 0, pb, scope->x, scope->y, scope->width, scope->height, FALSE);

	if (scope->mode == SCOPE_RENDER_LINES)
		{
		if (scope->horizontal)
			{
			gint w = scope->width / 2;
			for (i = 0; i < scope->points_count - 1; i++)
				{
				pixbuf_copy_line(scope->pixbuf, w - scope->points[i], scope->height + i,
						 w - scope->points[i+1], scope->height + i + 1,
						 pb, scope->x + w - scope->points[i], scope->y + i, FALSE);
				}
			}
		else
			{
			gint h = scope->height / 2;
			for (i = 0; i < scope->points_count - 1; i++)
				{
				pixbuf_copy_line(scope->pixbuf, i, scope->height + h - scope->points[i],
						 i + 1, scope->height + h - scope->points[i+1],
						 pb, scope->x + i, scope->y + h - scope->points[i], FALSE);
				}
                        }

		}
	else
		{
		if (scope->horizontal)
			{
			gint w = scope->width / 2;
			for (i = 0; i < scope->points_count; i++)
				{
				pixbuf_copy_point(scope->pixbuf, w + scope->points[i], scope->height + i,
						  pb, scope->x + w + scope->points[i], scope->y + i, FALSE);
				}
                        }
                else
                        {
			gint h = scope->height / 2;
			for (i = 0; i < scope->points_count; i++)
				{
				pixbuf_copy_point(scope->pixbuf, i, scope->height + h - scope->points[i],
						  pb, scope->x + i, scope->y + h - scope->points[i], FALSE);
				}
                        }

		}

	ui_display_render_area(ui, scope->x, scope->y, scope->width, scope->height, NULL);
}

static void scope_draw(gpointer data, const gchar *key, gint update, gint force, GdkPixbuf *pb, UIData *ui)
{
	ScopeData *scope = data;

	if (!scope) return;

	if (update)
		{
		ScopeCallbackData *cd;

		cd = ui_get_registered_callbacks(ui, key, type_id);
		if (cd && cd->status_get_func)
			{
			force = cd->status_get_func(scope, key, cd->status_get_data) || force;
			}
		if (cd && cd->mode_get_func)
			{
			cd->mode_get_func(scope, key, cd->mode_get_data);
			}
		}

	if (force)
		{
		scope_draw_real(scope, pb, ui);
		}
}

/*
 *-----------------------------
 * ui funcs
 *-----------------------------
 */

static void scope_back_set(gpointer data, GdkPixbuf *pb)
{
	ScopeData *scope = data;
	gint i;
	gint sh, sw;

	if (!scope) return;

	for (i = 0; i < 2; i++)
		{
		pixbuf_copy_area(pb, scope->x, scope->y, scope->pixbuf,
				 0, i * scope->height, scope->width, scope->height, FALSE);
		}

	sw = gdk_pixbuf_get_width(scope->overlay);
	sh = gdk_pixbuf_get_height(scope->overlay);

	if (scope->stretch)
		{
		pixbuf_copy_fill_alpha(scope->overlay, 0, 0, sw, sh,
				       scope->pixbuf, 0, 0, scope->width, scope->height * 2,
				       TRUE, 255);
		return;
		}
	pixbuf_copy_fill_alpha(scope->overlay, 0, 0, sw, sh / 2,
			       scope->pixbuf, 0, 0, scope->width, scope->height,
			       FALSE, 255);
	pixbuf_copy_fill_alpha(scope->overlay, 0, sh / 2, sw, sh / 2,
			       scope->pixbuf, 0, scope->height, scope->width, scope->height,
			       FALSE, 255);
}

static gint scope_get_geometry(gpointer widget, gint *x, gint *y, gint *w, gint *h)
{
	ScopeData *scope = widget;

	*x = scope->x;
	*y = scope->y;
	*w = scope->width;
	*h = scope->height;

	return TRUE;
}

static void scope_set_coord(gpointer widget, gint x, gint y)
{
	ScopeData *scope = widget;

	scope->x = x;
	scope->y = y;
}

static void scope_set_size(gpointer widget, gint dev_w, gint dev_h)
{
	ScopeData *scope = widget;
	gint w, h;

	if (!scope->dynamic) return;

	w = scope->width + dev_w;
	h = scope->height + dev_h;

	if (w < 4) w = 4;
	if (h < 4) h = 4;

	scope_sync_to_size(scope, w, h, scope->horizontal);
}

static WidgetData *scope_parse(SkinData *skin, GList *list, const gchar *skin_dir, const gchar *key, gint edit)
{
	WidgetData *wd = NULL;
	ScopeData *scope;
	gchar *filename;
	gint x, y;
	gint horizontal;
	gint dynamic;
	gint stretch;
	gint w, h;

	/* req */
	if (!key_list_read_int(list, "x", &x)) return NULL;
	if (!key_list_read_int(list, "y", &y)) return NULL;
	if (!key_list_read_int(list, "width", &w)) w = -1;
	if (!key_list_read_int(list, "height", &h)) h = -1;
	filename = key_list_read_path(list, "image", skin_dir);
	if (!filename) return NULL;

	/* opt */
	horizontal = key_list_read_bool(list, "horizontal");
	dynamic = key_list_read_bool(list, "dynamic");
	stretch = key_list_read_bool(list, "stretch");

	scope = scope_new_from_file(filename, x, y, horizontal,
				    w, h, dynamic, stretch);

	if (scope)
		{
		wd = scope_register(skin, scope, key, NULL);

		if (edit)
			{
			ui_widget_set_data(wd, "image", filename);
			}
		}

	g_free(filename);

	return wd;
}


/*
 *-----------------------------
 * register ui / app side
 *-----------------------------
 */

WidgetData *scope_register(SkinData *skin, ScopeData *scope, const gchar *key, const gchar *text_id)
{
	return skin_register_widget(skin, key, text_id, type_id, scope);
}

WidgetData *scope_register_to_skin(const gchar *key, gchar **data, gint x, gint y,
				   gint w, gint h, gint horizontal, gint dynamic, gint stretch,
				   SkinData *skin, const gchar *text_id)
{
	ScopeData *scope;

	scope = scope_new_from_data(data, x, y, horizontal, w, h, dynamic, stretch);
	return scope_register(skin, scope, key, text_id);
}

RegisterData *scope_register_key(const gchar *key, UIData *ui,
				 gint (*status_get_func)(ScopeData *scope, const gchar *key, gpointer data), gpointer status_get_data,
				 void (*mode_get_func)(ScopeData *scope, const gchar *key, gpointer data), gpointer mode_get_data)
{
	ScopeCallbackData *cd;

	scope_type_init();

	cd = g_new0(ScopeCallbackData, 1);

	cd->status_get_func = status_get_func;
	cd->status_get_data = status_get_data;

	cd->mode_get_func = mode_get_func;
	cd->mode_get_data = mode_get_data;

	return ui_register_key(ui, key, type_id, cd, sizeof(ScopeCallbackData));
}

/*
 *-----------------------------
 * app funcs
 *-----------------------------
 */

static void scope_value_set_cb(WidgetData *wd, gpointer data, GdkPixbuf *pb, UIData *ui)
{
	ScopeData *scope;
	ScopeSetData *msd = data;

	scope = wd->widget;

	if (scope_parse_buffer(scope, msd->buffer, msd->length))
		{
		scope_draw_real(scope, pb, ui);
		}
}

gint scope_value_set(const gchar *key, UIData *ui, short *buffer, gint length)
{
	ScopeSetData msd;

	msd.buffer = buffer;
	msd.length = length;

	return skin_widget_for_each_key(ui, key, type_id, scope_value_set_cb, &msd);
}

static void scope_reset_cb(WidgetData *wd, gpointer data, GdkPixbuf *pb, UIData *ui)
{
	ScopeData *scope;

	scope = wd->widget;

	if (scope_parse_buffer(scope, NULL, 0))
		{
		scope_draw_real(scope, pb, ui);
		}
}

gint scope_reset(const gchar *key, UIData *ui)
{
	return skin_widget_for_each_key(ui, key, type_id, scope_reset_cb, NULL);
}

static void scope_set_options_cb(WidgetData *wd, gpointer data, GdkPixbuf *pb, UIData *ui)
{
	ScopeData *scope;
	ScopeSetData *msd = data;

	scope = wd->widget;

	scope->mode = msd->mode;
	scope->scale = msd->scale;
}

gint scope_set_options(const gchar *key, UIData *ui, ScopeRenderType mode, gint scale)
{
	ScopeSetData msd;

	msd.mode = mode;
	msd.scale = CLAMP(scale, 1, 20);

	return skin_widget_for_each_key(ui, key, type_id, scope_set_options_cb, &msd);
}

gint scope_widget_set(ScopeData *scope, short value)
{
	gint i;
	gfloat fact;
	gint ret = FALSE;

	if (!scope) return FALSE;

	if (scope->horizontal)
		{
		fact = (gfloat)65536 / scope->width;
		}
	else
		{
		fact = (gfloat)65536 / scope->height;
		}

	for (i = 0; i < scope->points_count; i++)
		{
		gint new_val;

		new_val = (gint)((gfloat)value / fact);
		if (scope->points[i] != new_val)
			{
			scope->points[i] = new_val;
			ret = TRUE;
			}
		}

	return ret;
}

gint scope_widget_reset(ScopeData *scope)
{
        return scope_widget_set(scope, 0);
}

gint scope_widget_set_buffer(ScopeData *scope, short *buffer, gint length)
{
	if (!scope) return FALSE;

	return scope_parse_buffer(scope, buffer, length);
}


/*
 *-----------------------------
 * init
 *-----------------------------
 */

WidgetType scope_type_id(void)
{
	return type_id;
}

void scope_type_init(void)
{
	WidgetObjectData *od;

	if (type_id != -1) return;

	od = ui_widget_type_new("scope");
	type_id = od->type;

	od->func_draw = scope_draw;
	od->func_back = scope_back_set;
	od->func_free = scope_free_cb;

	od->func_get_geometry = scope_get_geometry;
	od->func_set_coord = scope_set_coord;
	od->func_set_size = scope_set_size;

	od->func_parse = scope_parse;

	scope_type_init_edit(od);
}


