/*
 * (SLIK) SimpLIstic sKin functions
 * (C) 2002 John Ellis
 *
 * Author: John Ellis
 *
 * This software is released under the GNU General Public License (GNU GPL).
 * Please read the included file COPYING for more information.
 * This software comes with no warranty of any kind, use at your own risk!
 */

#include "ui2_includes.h"
#include "ui2_typedefs.h"
#include "ui2_spectrum.h"
#include "ui2_spectrum_edit.h"

#include "ui2_display.h"
#include "ui2_main.h"
#include "ui2_parse.h"
#include "ui2_skin.h"
#include "ui2_util.h"
#include "ui2_widget.h"
#include "ui_pixbuf_ops.h"

#include <math.h>

typedef struct _SpectrumCallbackData SpectrumCallbackData;
struct _SpectrumCallbackData
{
	gint (*status_get_func)(SpectrumData *spectrum, const gchar *key, gpointer data);
	gpointer status_get_data;
	void (*mode_get_func)(SpectrumData *spectrum, const gchar *key, gpointer data);
	gpointer mode_get_data;
};

typedef struct _SpectrumSetData SpectrumSetData;
struct _SpectrumSetData
{
	gint32 *source;
	gfloat fall_percent;
	SpectrumPeakType peak_mode;
	gfloat value;
};

static WidgetType type_id = -1;


/*
 *-----------------------------
 * sync internal data (arrays, etc. + only public for access by editor)
 *-----------------------------
 */

void spectrum_sync_band_count(SpectrumData *spectrum, gint bands)
{
	if (!spectrum) return;

	spectrum->bands = bands;

	if (spectrum->one_image)
		{
		if (spectrum->horizontal)
			{
			spectrum->width = gdk_pixbuf_get_width(spectrum->overlay);
			spectrum->height = gdk_pixbuf_get_height(spectrum->overlay) / 3 * bands;
			spectrum->band_size = gdk_pixbuf_get_height(spectrum->overlay) / 3;
			}
		else
			{
			spectrum->width = gdk_pixbuf_get_width(spectrum->overlay) * bands;
			spectrum->height = gdk_pixbuf_get_height(spectrum->overlay) / 3;
			spectrum->band_size = gdk_pixbuf_get_width(spectrum->overlay);
			}
		}
	else
		{
		spectrum->width = gdk_pixbuf_get_width(spectrum->overlay);
		spectrum->height = gdk_pixbuf_get_height(spectrum->overlay) / 3;
		if (spectrum->horizontal)
			{
			spectrum->band_size = spectrum->height / bands;
			}
		else
			{
			spectrum->band_size = spectrum->width / bands;
			}
		}
	if (spectrum->horizontal)
		{
		spectrum->segment_size = spectrum->width / spectrum->segments;
		}
	else
		{
		spectrum->segment_size = spectrum->height / spectrum->segments;
		}

	g_free(spectrum->values);
	g_free(spectrum->peak_values);
	spectrum->values = g_malloc(sizeof(float) * spectrum->bands);
	spectrum->peak_values = g_malloc(sizeof(float) * spectrum->bands);

	if (spectrum->pixbuf) gdk_pixbuf_unref(spectrum->pixbuf);
	spectrum->pixbuf = gdk_pixbuf_new(GDK_COLORSPACE_RGB, gdk_pixbuf_get_has_alpha(spectrum->overlay),
					  8, spectrum->width, spectrum->height * 3);
}

/*
 *-----------------------------
 * new / free
 *-----------------------------
 */

SpectrumData *spectrum_new(GdkPixbuf *pb, gint x, gint y, gint segments, gint bands,
			   gint horizontal, gint one_image, gint dynamic, gint slide)
{
	SpectrumData *spectrum;

	spectrum_type_init();

	util_size(&x);
	util_size(&y);

	if (!pb) return NULL;
	if (segments < 1) segments = 1;
	if (bands < 1) bands = 1;
	if (bands > 16) bands = 16;

	if (dynamic && !one_image)
		{
		dynamic = FALSE;
		printf("spectrum warning: dynamic specified for static image, disabled\n");
		}

	spectrum = g_new0(SpectrumData, 1);

	spectrum->overlay = util_size_pixbuf(pb, TRUE);

	spectrum->segments = segments;
	spectrum->x = x;
	spectrum->y = y;
	spectrum->horizontal = horizontal;
	spectrum->slide = slide;

	spectrum->peak_mode = SPECTRUM_PEAK_OFF;
	spectrum->peak_fall = 0.0;

	spectrum->one_image = one_image;
	spectrum->dynamic = dynamic;

	spectrum->values = NULL;
	spectrum->peak_values = NULL;
	spectrum->pixbuf = NULL;

	spectrum_sync_band_count(spectrum, bands);

	spectrum->dynamic_size = spectrum->bands * spectrum->band_size;

	return spectrum;
}

SpectrumData *spectrum_new_from_data(gchar **data, gint x, gint y, gint segments, gint bands,
				     gint horizontal, gint one_image, gint dynamic, gint slide)
{
	GdkPixbuf *pb;

	pb = gdk_pixbuf_new_from_xpm_data((const char **)data);

	return spectrum_new(pb, x, y, segments, bands, horizontal, one_image, dynamic, slide);
}

SpectrumData *spectrum_new_from_file(const gchar *file, gint x, gint y, gint segments, gint bands,
				     gint horizontal, gint one_image, gint dynamic, gint slide)
{
	GdkPixbuf *pb;

	pb = gdk_pixbuf_new_from_file(file, NULL);

	return spectrum_new(pb, x, y, segments, bands, horizontal, one_image, dynamic, slide);
}

void spectrum_free(SpectrumData *spectrum)
{
	if (!spectrum) return;
	if (spectrum->pixbuf) gdk_pixbuf_unref(spectrum->pixbuf);
	if (spectrum->overlay) gdk_pixbuf_unref(spectrum->overlay);
	g_free(spectrum->values);
	g_free(spectrum->peak_values);
	g_free(spectrum);
}

static void spectrum_free_cb(gpointer data)
{
	spectrum_free((SpectrumData *)data);
}

/*
 *-----------------------------
 * fft parser (source must have SPECTRUM_BUFFER_SIZE (256) elements)
 *-----------------------------
 */

static const gint spectrum_points[] = {0, 1, 2, 3, 5, 7, 10, 14, 20,
				       28, 40, 54, 74, 101, 137, 187, 255};
#define SPECTRUM_POINTS_COUNT 16

static void spectrum_parse_values(gfloat *response, gint32 *source, gint bands)
{
	gint i;

	for (i = 0; i < bands; i++)
		{
		gfloat f;
		gint n;
		gint32 p = 0;
		gint l, h;

		l = spectrum_points[SPECTRUM_POINTS_COUNT / bands * i];
		h = spectrum_points[SPECTRUM_POINTS_COUNT / bands * (i + 1)];

		for (n = l; n < h; n++)
			{
			if (source[n] > p) p = source[n];
			}

		p >>= 7;        /* shift to 24-bit number, max 16.7mil, remember p is signed */

		if (p != 0)
			{
			f = log(p) / 7.22471987;        /* 7.2 is log(16mil) */
			}
		else
			{
			f = 0.0;
			}
		response[i] = f;
		}
}

/*
 *-----------------------------
 * draw
 *-----------------------------
 */

static void spectrum_draw_layer(SpectrumData *spectrum, GdkPixbuf *pb, gfloat *values, gint offset, gint fill)
{
	gint i;

	if (spectrum->horizontal)
		{
		for (i = 0; i < spectrum->bands; i++)
			{
			gint x, y, w, h;

			w = (((gint)(values[i] * (gfloat)spectrum->segments)) * spectrum->segment_size);
			y = spectrum->height - ((i + 1) * spectrum->band_size);
			h = spectrum->band_size;

			if (fill || spectrum->slide)
				{
				x = 0;
				}
			else
				{
				x = w - spectrum->segment_size;
				w = spectrum->segment_size;
				}
			if (w)
				{
				gint p;

				if (spectrum->slide)
					{
					p = spectrum->width - w;
					}
				else
					{
					p = x;
					}
				pixbuf_copy_area_alpha(spectrum->pixbuf, p, offset * spectrum->height + y,
						       pb, spectrum->x + x, spectrum->y + y, w, h, 255);
				}
			}
		}
	else
		{
		for (i = 0; i < spectrum->bands; i++)
			{
			gint x, y, w, h;

			x = i * spectrum->band_size;
			y = spectrum->height - (((gint)(values[i] * (gfloat)spectrum->segments)) * spectrum->segment_size);
			w = spectrum->band_size;

			if (y == spectrum->height)
				{
				h = 0;
				}
			else if (fill || spectrum->slide)
				{
				h = spectrum->height - y;
				}
			else
				{
				h = spectrum->segment_size;
				}
			if (h)
				{
				gint p;

				if (spectrum->slide)
					{
					p = offset * spectrum->height;
					}
				else
					{
					p = offset * spectrum->height + y;
					}
				pixbuf_copy_area_alpha(spectrum->pixbuf, x, p,
						       pb, spectrum->x + x, spectrum->y + y, w, h, 255);
				}
			}
		}
}

static void spectrum_draw_real(SpectrumData *spectrum, GdkPixbuf *pb, UIData *ui)
{
	if (!spectrum) return;

	/* FIXME! flip and horizontal not implemented yet! */

	pixbuf_copy_area(spectrum->pixbuf, 0, 0, pb,
			 spectrum->x, spectrum->y, spectrum->width, spectrum->height, FALSE);

	spectrum_draw_layer(spectrum, pb, spectrum->values, 1, TRUE);
	if (spectrum->peak_mode != SPECTRUM_PEAK_OFF)
		{
		spectrum_draw_layer(spectrum, pb, spectrum->peak_values, 2, FALSE);
		}

	ui_display_render_area(ui, spectrum->x, spectrum->y, spectrum->width, spectrum->height, NULL);
}

static gint spectrum_value_sync(SpectrumData *spectrum, gfloat *new_values, gfloat fall_percent)
{
	gint ret = FALSE;
	gfloat fall;
	gint i;

	fall = fall_percent * spectrum->peak_fall;

	for (i = 0 ; i < spectrum->bands; i++)
		{
		gfloat old_peak;
		gfloat new_peak;
		gfloat new_val;

		old_peak = spectrum->peak_values[i];

		new_val = (new_values) ? new_values[i] : 0.0;

		switch (spectrum->peak_mode)
			{
			case SPECTRUM_PEAK_OFF:
			default:
				new_peak = 0.0;
				break;
			case SPECTRUM_PEAK_ACTIVE:
				if (new_val >= old_peak)
					{
					new_peak = new_val;
					}
				else
					{
					new_peak = old_peak - fall;
					if (new_peak < 0.0) new_peak = 0.0;
					} 
				break;
			case SPECTRUM_PEAK_SMOOTH:
				if (new_val < old_peak)
					{
					new_peak = old_peak - fall;
					if (new_peak < 0.0) new_peak = 0.0;
					}
				else if (new_val > old_peak)
					{
					new_peak = old_peak + fall;
					if (new_peak > 1.0) new_peak = 1.0;
					}
				else
					{
					new_peak = old_peak;
					}
				break;
			}

		if (spectrum->values[i] != new_val || old_peak != new_peak)
			{
			spectrum->values[i] = new_val;
			spectrum->peak_values[i] = new_peak;
			ret = TRUE;
			}
		}

	return ret;
}

static void spectrum_draw(gpointer data, const gchar *key, gint update, gint force, GdkPixbuf *pb, UIData *ui)
{
	SpectrumData *spectrum = data;

	if (!spectrum) return;

	if (update)
		{
		SpectrumCallbackData *cd;

		cd = ui_get_registered_callbacks(ui, key, type_id);
		if (cd && cd->status_get_func)
			{
			force = cd->status_get_func(spectrum, key, cd->status_get_data) || force;
			}
		if (cd && cd->mode_get_func)
			{
			cd->mode_get_func(spectrum, key, cd->mode_get_data);
			}
		}

	if (force)
		{
		spectrum_draw_real(spectrum, pb, ui);
		}
}

/*
 *-----------------------------
 * ui funcs
 *-----------------------------
 */

static void spectrum_back_set(gpointer data, GdkPixbuf *pb)
{
	SpectrumData *spectrum = data;
	gint i;

	if (!spectrum) return;

	pixbuf_copy_area(pb, spectrum->x, spectrum->y, spectrum->pixbuf,
			 0, 0, spectrum->width, spectrum->height, FALSE);

	if (spectrum->one_image)
		{
		gint w, h, t;

		w = gdk_pixbuf_get_width(spectrum->overlay);
		t = gdk_pixbuf_get_height(spectrum->overlay);
		h = t / 3;

		if (spectrum->horizontal)
			{
			for (i = 0; i < spectrum->bands; i++)
				{
				pixbuf_copy_area_alpha(spectrum->overlay, 0, 0,
						       spectrum->pixbuf, 0, i * h,
						       w, h, 255);
				pixbuf_copy_area(spectrum->overlay, 0, h,
						 spectrum->pixbuf, 0, (spectrum->bands * h) + (i * h),
						 w, h, TRUE);
				pixbuf_copy_area(spectrum->overlay, 0, h + h,
						 spectrum->pixbuf, 0, (spectrum->bands * h * 2) + (i * h),
						 w, h, TRUE);
				}
			}
		else
			{
			for (i = 0; i < spectrum->bands; i++)
				{
				pixbuf_copy_area_alpha(spectrum->overlay, 0, 0,
						       spectrum->pixbuf, i * w, 0,
						       w, h, 255);
				pixbuf_copy_area(spectrum->overlay, 0, h,
						 spectrum->pixbuf, i * w, h,
						 w, t - h, TRUE);
				}
			}
		}
	else
		{
		pixbuf_copy_area_alpha(spectrum->overlay, 0, 0,
				       spectrum->pixbuf, 0, 0,
				       spectrum->width, spectrum->height, 255);
		pixbuf_copy_area(spectrum->overlay, 0, spectrum->height,
				 spectrum->pixbuf, 0, spectrum->height,
				 spectrum->width, spectrum->height * 2, TRUE);
		}
}

static gint spectrum_get_geometry(gpointer widget, gint *x, gint *y, gint *w, gint *h)
{
	SpectrumData *spectrum = widget;

	*x = spectrum->x;
	*y = spectrum->y;
	*w = spectrum->width;
	*h = spectrum->height;

	return TRUE;
}

static void spectrum_set_coord(gpointer widget, gint x, gint y)
{
	SpectrumData *spectrum = widget;

	if (spectrum->horizontal)
		{
		if (spectrum->y != y) spectrum->dynamic_size = spectrum->height;
		}
	else
		{
		if (spectrum->x != x) spectrum->dynamic_size = spectrum->width;
		}

	spectrum->x = x;
	spectrum->y = y;
}

static void spectrum_set_size(gpointer widget, gint dev_w, gint dev_h)
{
	SpectrumData *spectrum = widget;
	gint old_count;
	gint new_count;
	gint dev_size;

	if (!spectrum->one_image || !spectrum->dynamic) return;

	if (spectrum->horizontal)
		{
		dev_size = dev_h;
		}
	else
		{
		dev_size = dev_w;
		}

	spectrum->dynamic_size += dev_size;
	if (spectrum->dynamic_size < spectrum->band_size) spectrum->dynamic_size = spectrum->band_size;

	old_count = spectrum->bands;
	new_count = spectrum->dynamic_size / spectrum->band_size;

	if (new_count > 16) new_count = 16;
	if (new_count == old_count) return;

	spectrum_sync_band_count(spectrum, new_count);
}

static WidgetData *spectrum_parse(SkinData *skin, GList *list, const gchar *skin_dir, const gchar *key, gint edit)
{
	WidgetData *wd = NULL;
	SpectrumData *spectrum;
	gchar *filename;
	gint x, y;
	gint segments;
	gint bands;
	gint horizontal;
	gint one_image;
	gint dynamic;
	gint slide;

	/* req */
	if (!key_list_read_int(list, "x", &x)) return NULL;
	if (!key_list_read_int(list, "y", &y)) return NULL;
	if (!key_list_read_int(list, "segments", &segments) || segments == 0) return NULL;
	if (!key_list_read_int(list, "bands", &bands) || bands == 0) return NULL;
	filename = key_list_read_path(list, "image", skin_dir);
	if (!filename) return NULL;

	/* opt */
	horizontal = key_list_read_bool(list, "horizontal");
	one_image = key_list_read_bool(list, "single_band");
	dynamic = key_list_read_bool(list, "dynamic");
	slide = key_list_read_bool(list, "slide");

	spectrum = spectrum_new_from_file(filename, x, y, segments, bands,
					  horizontal, one_image, dynamic, slide);

	if (spectrum)
		{
		wd = spectrum_register(skin, spectrum, key, NULL);

		if (edit)
			{
			ui_widget_set_data(wd, "image", filename);
			}
		}

	g_free(filename);

	return wd;
}

/*
 *-----------------------------
 * register ui / app side
 *-----------------------------
 */

WidgetData *spectrum_register(SkinData *skin, SpectrumData *spectrum, const gchar *key, const gchar *text_id)
{
	return skin_register_widget(skin, key, text_id, type_id, spectrum);
}

WidgetData *spectrum_register_to_skin(const gchar *key, gchar **data, gint x, gint y,
				      gint segments, gint bands,
				      gint horizontal, gint one_image, gint dynamic, gint slide,
				      SkinData *skin, const gchar *text_id)
{
	SpectrumData *spectrum;

	spectrum = spectrum_new_from_data(data, segments, bands, x, y,
					  horizontal, one_image, dynamic, slide);
	return spectrum_register(skin, spectrum, key, text_id);
}

RegisterData *spectrum_register_key(const gchar *key, UIData *ui,
				 gint (*status_get_func)(SpectrumData *spectrum, const gchar *key, gpointer data), gpointer status_get_data,
				 void (*mode_get_func)(SpectrumData *spectrum, const gchar *key, gpointer data), gpointer mode_get_data)
{
	SpectrumCallbackData *cd;

	spectrum_type_init();

	cd = g_new0(SpectrumCallbackData, 1);

	cd->status_get_func = status_get_func;
	cd->status_get_data = status_get_data;

	cd->mode_get_func = mode_get_func;
	cd->mode_get_data = mode_get_data;

	return ui_register_key(ui, key, type_id, cd, sizeof(SpectrumCallbackData));
}

/*
 *-----------------------------
 * app funcs
 *-----------------------------
 */

static void spectrum_value_set_cb(WidgetData *wd, gpointer data, GdkPixbuf *pb, UIData *ui)
{
	SpectrumData *spectrum;
	SpectrumSetData *msd = data;
	gfloat *new_vals;

	spectrum = wd->widget;

	if (msd->source)
		{
		new_vals = g_malloc(sizeof(gfloat) * spectrum->bands);
		spectrum_parse_values(new_vals, msd->source, spectrum->bands);
		}
	else
		{
		new_vals = NULL;
		}

	if (spectrum_value_sync(spectrum, new_vals, msd->fall_percent))
		{
		spectrum_draw_real(spectrum, pb, ui);
		}

	g_free(new_vals);
}

/* buffer is a pointer to gint32 array with 256 elements */
gint spectrum_value_set(const gchar *key, UIData *ui, gint32 *buffer, gfloat fall_percent)
{
	SpectrumSetData msd;

	msd.source = buffer;
	msd.fall_percent = CLAMP(fall_percent, 0.0, 1.0);

	return skin_widget_for_each_key(ui, key, type_id, spectrum_value_set_cb, &msd);
}

static void spectrum_value_all_one_cb(WidgetData *wd, gpointer data, GdkPixbuf *pb, UIData *ui)
{
	SpectrumData *spectrum;
	SpectrumSetData *msd = data;

	spectrum = wd->widget;

	if (spectrum_widget_set(spectrum, msd->value))
		{
		spectrum_draw_real(spectrum, pb, ui);
		}
}

gint spectrum_value_all_one(const gchar *key, UIData *ui, gfloat value, gfloat fall_percent)
{
	SpectrumSetData msd;

	msd.value = value;
	msd.fall_percent = CLAMP(fall_percent, 0.0, 1.0);

	return skin_widget_for_each_key(ui, key, type_id, spectrum_value_all_one_cb, &msd);
}

static void spectrum_reset_cb(WidgetData *wd, gpointer data, GdkPixbuf *pb, UIData *ui)
{
	SpectrumData *spectrum;

	spectrum = wd->widget;

	if (spectrum_widget_reset(spectrum))
		{
		spectrum_draw_real(spectrum, pb, ui);
		}
}

gint spectrum_reset(const gchar *key, UIData *ui)
{
	return skin_widget_for_each_key(ui, key, type_id, spectrum_reset_cb, NULL);
}

static void spectrum_set_options_cb(WidgetData *wd, gpointer data, GdkPixbuf *pb, UIData *ui)
{
	SpectrumData *spectrum;
	SpectrumSetData *msd = data;

	spectrum = wd->widget;

	spectrum->peak_mode = msd->peak_mode;
	spectrum->peak_fall = msd->fall_percent;
}

gint spectrum_set_options(const gchar *key, UIData *ui, SpectrumPeakType mode, gfloat fall_speed)
{
	SpectrumSetData msd;

	msd.peak_mode = mode;
	msd.fall_percent = CLAMP(fall_speed, 0.0, 1.0);

	return skin_widget_for_each_key(ui, key, type_id, spectrum_set_options_cb, &msd);
}

/* utils used when status is requested */
gint spectrum_widget_set(SpectrumData *spectrum, gfloat value)
{
	gint i;
	gint ret = FALSE;

	if (!spectrum) return FALSE;

	for (i = 0; i < spectrum->bands; i++)
		{
		if (spectrum->values[i] != value || spectrum->peak_values[i] > 0.0) ret = TRUE;

		spectrum->values[i] = value;
		spectrum->peak_values[i] = 0.0;
		}

	return ret;
}

gint spectrum_widget_reset(SpectrumData *spectrum)
{
	return spectrum_widget_set(spectrum, 0.0);
}

gint spectrum_widget_set_buffer(SpectrumData *spectrum, gint32 *buffer)
{
	gfloat *new_vals;
	gint ret;

	if (!spectrum) return FALSE;

	if (buffer)
		{
		new_vals = g_malloc(sizeof(gfloat) * spectrum->bands);
		spectrum_parse_values(new_vals, buffer, spectrum->bands);
		}
	else
		{
		new_vals = NULL;
		}

	ret = spectrum_value_sync(spectrum, new_vals, 0.0);

	g_free(new_vals);

	return ret;
}

/*
 *-----------------------------
 * init
 *-----------------------------
 */

WidgetType spectrum_type_id(void)
{
	return type_id;
}

void spectrum_type_init(void)
{
	WidgetObjectData *od;

	if (type_id != -1) return;

	od = ui_widget_type_new("spectrum");
	type_id = od->type;

	od->func_draw = spectrum_draw;
	od->func_back = spectrum_back_set;
	od->func_free = spectrum_free_cb;

	od->func_get_geometry = spectrum_get_geometry;
	od->func_set_coord = spectrum_set_coord;
	od->func_set_size = spectrum_set_size;

	od->func_parse = spectrum_parse;

	spectrum_type_init_edit(od);
}


