/*
 * (SLIK) SimpLIstic sKin functions
 * (C) 2002 John Ellis
 *
 * Author: John Ellis
 *
 * This software is released under the GNU General Public License (GNU GPL).
 * Please read the included file COPYING for more information.
 * This software comes with no warranty of any kind, use at your own risk!
 */

#include "ui2_includes.h"
#include "ui2_typedefs.h"

#include "ui2_spectrum.h"
#include "ui2_spectrum_edit.h"

#include "ui2_editor.h"
#include "ui2_widget.h"
#include "ui_fileops.h"
#include "ui_tabcomp.h"
#include "ui_utildlg.h"


typedef struct _SpectrumDetail SpectrumDetail;
struct _SpectrumDetail
{
	SpectrumData *spectrum;
	GtkWidget *segments_spin;
	GtkWidget *bands_spin;
	GtkWidget *horizontal_toggle;
	GtkWidget *slide_toggle;

	GtkWidget *one_image_toggle;
	GtkWidget *dynamic_toggle;

	WidgetData *wd;
	UIData *ui;
};

typedef struct _SpectrumListData SpectrumListData;
struct _SpectrumListData
{
	gchar *key;
	gchar *text_id;
	gchar *data;

	gchar *image;
	gint segments;
	gint bands;
	gint horizontal;
	gint slide;
	gint one_image;
	gint dynamic;
};

typedef struct _SpectrumPage SpectrumPage;
struct _SpectrumPage
{
	GtkWidget *key_entry;
	GtkWidget *text_id_entry;
	GtkWidget *data_entry;

	GtkWidget *image_entry;
	GtkWidget *segment_spin;
	GtkWidget *bands_spin;
	GtkWidget *horizontal_toggle;
	GtkWidget *slide_toggle;

	GtkWidget *one_image_toggle;
	GtkWidget *dynamic_toggle;

	GtkWidget *clist;
	GtkWidget *pixmap;

	EditData *ed;
};


static GdkPixbuf *spectrum_get_pixbuf(gpointer data)
{
	SpectrumData *spectrum = data;
	return spectrum->overlay;
}

static void spectrum_edit_write(FILE *f, WidgetData *wd, SkinData *skin, const gchar *dir)
{
	SpectrumData *spectrum = wd->widget;
	gchar *image;

	image = ui_edit_copy_unique_file(ui_widget_get_data(wd, "image"),
					 dir, spectrum->overlay, "spectrum_", wd->key);

	if (image) ui_edit_widget_set_path_key(wd, "image", dir, image);

	ui_edit_write_section(f, "spectrum", wd->key);

	ui_edit_write_key_char(f, "image", image);
	ui_edit_write_key_int(f, "x", spectrum->x);
	ui_edit_write_key_int(f, "y", spectrum->y);
	ui_edit_write_key_int(f, "segments", spectrum->segments);
	ui_edit_write_key_int(f, "bands", spectrum->bands);
	ui_edit_write_key_bool(f, "horizontal", spectrum->horizontal);
	ui_edit_write_key_bool(f, "slide", spectrum->slide);

	ui_edit_write_key_bool(f, "single_band", spectrum->one_image);
	ui_edit_write_key_bool(f, "dynamic", (spectrum->dynamic && spectrum->one_image));

	g_free(image);
}

static SpectrumListData *spectrum_edit_list_find(GList *list, const gchar *image)
{
	GList *work;
	work = list;
	while (work)
		{
		SpectrumListData *sd = work->data;
		if (strcmp(image, sd->image) == 0) return sd;
		work = work->next;
		}
	return NULL;
}

static gpointer spectrum_edit_read(UIData *ui, WidgetData *wd, GList *list)
{
	SpectrumData *spectrum = wd->widget;
	SpectrumListData *sd;
	const gchar *image;

	image = ui_widget_get_data(wd, "image");

	if (!image || spectrum_edit_list_find(list, image)) return NULL;

	sd = g_new0(SpectrumListData, 1);
	sd->image = g_strdup(image);
	sd->segments = spectrum->segments;
	sd->bands = spectrum->bands;
	sd->horizontal = spectrum->horizontal;
	sd->slide = spectrum->slide;
	sd->one_image = spectrum->one_image;
	sd->dynamic = spectrum->dynamic;

	sd->key = g_strdup(wd->key);
	sd->data = g_strdup(ui_widget_get_data(wd, "data"));
	sd->text_id = g_strdup(wd->text_id);

	return sd;
}

static void spectrum_edit_props_segment_cb(GtkAdjustment *adjustment, gpointer data)
{
	SpectrumDetail *sd = data;
	gint new_count;
	gint max_size;

	new_count = ui_edit_spin_get(sd->segments_spin);

	if (sd->spectrum->horizontal)
		{
		max_size = sd->spectrum->width;
		}
	else
		{
		max_size = sd->spectrum->height;
		}

	if (new_count > max_size)
		{
		new_count = max_size;
		ui_edit_spin_set_blocking(sd->segments_spin, new_count, sd);
		}

	if (sd->spectrum->segments == new_count) return;

	sd->spectrum->segments = new_count;
	sd->spectrum->segment_size = max_size / new_count;
}

static void spectrum_edit_props_bands_cb(GtkAdjustment *adjustment, gpointer data)
{
	SpectrumDetail *sd = data;
	SpectrumData *spectrum;
	gint old_w;
	gint new_count;
	gint max_size;

	spectrum = sd->spectrum;

	new_count = ui_edit_spin_get(sd->bands_spin);

	if (spectrum->one_image)
		{
		max_size = 16;
		}
	else if (spectrum->horizontal)
		{
		max_size = spectrum->height;
		}
	else
		{
		max_size = spectrum->width;
		}

	if (new_count > max_size || new_count > 16)
		{
		if (new_count > max_size) new_count = max_size;
		if (new_count > 16) new_count = 16;
		ui_edit_spin_set_blocking(sd->bands_spin, new_count, sd);
		}
	if (new_count > spectrum->bands &&
	    ((!spectrum->horizontal && spectrum->x + (new_count * spectrum->band_size) >= sd->ui->skin->width) ||
	    (spectrum->horizontal && spectrum->y + (new_count * spectrum->band_size) >= sd->ui->skin->height)) )
		{
		ui_edit_spin_set_blocking(sd->bands_spin, spectrum->bands, sd);
		return;
		}


	if (spectrum->bands == new_count) return;

	old_w = spectrum->width;

	ui_edit_widget_draw_highlight(sd->ui, sd->wd, FALSE);

	spectrum_sync_band_count(spectrum, new_count);
	spectrum_widget_reset(spectrum);

	spectrum->dynamic_size = spectrum->bands * spectrum->band_size;
	
	ui_edit_widget_resync(sd->ui, sd->wd, TRUE, spectrum->width - old_w, 0);
}

static void spectrum_edit_props_dynamic_cb(GtkWidget *widget, gpointer data)
{
	SpectrumDetail *sd = data;

	sd->spectrum->dynamic = ui_edit_toggle_get(sd->dynamic_toggle);
}

static void spectrum_edit_props_slide_cb(GtkWidget *widget, gpointer data)
{
	SpectrumDetail *sd = data;

	sd->spectrum->slide = ui_edit_toggle_get(sd->slide_toggle);
}

static gpointer spectrum_edit_props(UIData *ui, WidgetData *wd, GtkWidget *vbox, gpointer detail)
{
	SpectrumData *spectrum = wd->widget;
	SpectrumDetail *sd = detail;

	if (!sd)
		{
		GtkObject *adj;

		sd = g_new0(SpectrumDetail, 1);

		sd->segments_spin = ui_edit_spin_new(vbox, _("Segments:"), 1, 2000, &adj);
		gtk_signal_connect(GTK_OBJECT(adj), "value_changed",
				   GTK_SIGNAL_FUNC(spectrum_edit_props_segment_cb), sd);
		sd->bands_spin = ui_edit_spin_new(vbox, _("Bands:"), 1, 16, &adj);
		gtk_signal_connect(GTK_OBJECT(adj), "value_changed",
				   GTK_SIGNAL_FUNC(spectrum_edit_props_bands_cb), sd);
		sd->horizontal_toggle = ui_edit_toggle_new(vbox, _("Horizontal"));
		gtk_widget_set_sensitive(sd->horizontal_toggle, FALSE);

		sd->slide_toggle = ui_edit_toggle_new(vbox, _("Slide"));
		gtk_signal_connect(GTK_OBJECT(sd->slide_toggle), "clicked",
				   GTK_SIGNAL_FUNC(spectrum_edit_props_slide_cb), sd);

		sd->one_image_toggle = ui_edit_toggle_new(vbox, _("Single image"));
		gtk_widget_set_sensitive(sd->one_image_toggle, FALSE);
		sd->dynamic_toggle = ui_edit_toggle_new(vbox, _("Dynamic size"));
		gtk_widget_set_sensitive(sd->dynamic_toggle, spectrum->one_image);
		gtk_signal_connect(GTK_OBJECT(sd->dynamic_toggle), "clicked",
				   GTK_SIGNAL_FUNC(spectrum_edit_props_dynamic_cb), sd);
		}

	sd->spectrum = spectrum;
	sd->wd = wd;
	sd->ui = ui;

	ui_edit_spin_set_blocking(sd->segments_spin, spectrum->segments, sd);
	ui_edit_spin_set_blocking(sd->bands_spin, spectrum->bands, sd);
	ui_edit_toggle_set(sd->horizontal_toggle, spectrum->horizontal);
	ui_edit_toggle_set(sd->slide_toggle, spectrum->slide);

	ui_edit_toggle_set(sd->one_image_toggle, spectrum->one_image);
	ui_edit_toggle_set(sd->dynamic_toggle, spectrum->dynamic);

	return sd;
}

static void spectrum_edit_free(gpointer data)
{
	SpectrumListData *sd = data;
	g_free(sd->key);
	g_free(sd->text_id);
	g_free(sd->data);
	g_free(sd->image);
	g_free(sd);
}

static void spectrum_edit_page_add_cb(GtkWidget *widget, gpointer data)
{
	SpectrumPage *pd = data;
	SpectrumData *spectrum;
	const gchar *key;
	const gchar *text_id;
	const gchar *image;

	key = ui_edit_entry_get(pd->key_entry);
	text_id = ui_edit_entry_get(pd->text_id_entry);
	image = ui_edit_entry_get(pd->image_entry);

	if (!key || !image || !isfile(image))
		{
		warning_dialog(_("Spectrum error"), _("Spectrum must contain a key and valid image."));
		return;
		}

	spectrum = spectrum_new_from_file(image, 0, 0,
				    ui_edit_spin_get(pd->segment_spin),
				    ui_edit_spin_get(pd->bands_spin),
				    ui_edit_toggle_get(pd->horizontal_toggle),
				    ui_edit_toggle_get(pd->one_image_toggle),
				    ui_edit_toggle_get(pd->dynamic_toggle),
				    ui_edit_toggle_get(pd->slide_toggle));
	if (spectrum)
		{
		WidgetData *wd;
		wd = spectrum_register(pd->ed->ui->skin, spectrum, key, text_id);
		ui_edit_widget_add_finish(pd->ed, wd, image, ui_edit_entry_get(pd->data_entry));
		}

	tab_completion_append_to_history(pd->image_entry, image);
}

static void spectrum_edit_page_one_image_cb(GtkWidget *widget, gpointer data)
{
	SpectrumPage *pd = data;

	gtk_widget_set_sensitive(pd->dynamic_toggle, ui_edit_toggle_get(pd->one_image_toggle));
}

static void spectrum_edit_page_sync(SpectrumPage *pd, SpectrumListData *sd)
{
	if (!sd) return;

	ui_edit_entry_set(pd->key_entry, sd->key);
	ui_edit_entry_set(pd->text_id_entry, sd->text_id);
	ui_edit_entry_set(pd->data_entry, sd->data);

	ui_edit_entry_set(pd->image_entry, sd->image);

	ui_edit_spin_set(pd->segment_spin, sd->segments);
	ui_edit_spin_set(pd->bands_spin, sd->bands);
	ui_edit_toggle_set(pd->horizontal_toggle, sd->horizontal);
	ui_edit_toggle_set(pd->slide_toggle, sd->slide);

	ui_edit_toggle_set(pd->one_image_toggle, sd->one_image);
	ui_edit_toggle_set(pd->dynamic_toggle, sd->dynamic);
	gtk_widget_set_sensitive(pd->dynamic_toggle, sd->one_image);
}

static void spectrum_edit_page_clist_cb(GtkWidget *clist, gint row, gint col, GdkEvent *event, gpointer data)
{
	SpectrumPage *pd = data;
	SpectrumListData *md;

	md = gtk_clist_get_row_data(GTK_CLIST(clist), row);
	spectrum_edit_page_sync(pd, md);
}

static void spectrum_edit_page_destroy_cb(GtkWidget *widget, gpointer data)
{
	SpectrumPage *pd = data;

	g_free(pd);
}

static GtkWidget *spectrum_edit_page_new(EditData *ed)
{
	GtkWidget *hbox;
	GtkWidget *vbox;
	GtkWidget *spectrum;
	SpectrumPage *pd;
	gchar *titles[] = { _("Image"), _("Key"), _("Segments"), _("Bands"), NULL };

	pd = g_new0(SpectrumPage, 1);
	pd->ed = ed;

	hbox = gtk_hbox_new(FALSE, 5);
	gtk_container_set_border_width(GTK_CONTAINER(hbox), 5);
	gtk_object_set_data(GTK_OBJECT(hbox), "page", pd);
	gtk_signal_connect(GTK_OBJECT(hbox), "destroy",
			   GTK_SIGNAL_FUNC(spectrum_edit_page_destroy_cb), pd);

	vbox = gtk_vbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(hbox), vbox, FALSE, FALSE, 0);
	gtk_widget_show(vbox);

	pd->key_entry = ui_edit_key_entry_new(vbox, ed->ui, spectrum_type_id());
	pd->data_entry = ui_edit_entry_new(vbox, _("Data:"));
	pd->text_id_entry = ui_edit_entry_new(vbox, _("Text id:"));

	pd->segment_spin = ui_edit_spin_new(vbox, _("Segments:"), 1, 1000, NULL);
	pd->bands_spin = ui_edit_spin_new(vbox, _("Bands:"), 1, 16, NULL);
	pd->horizontal_toggle = ui_edit_toggle_new(vbox, _("Horizontal"));
	pd->slide_toggle = ui_edit_toggle_new(vbox, _("Slide"));
	pd->one_image_toggle = ui_edit_toggle_new(vbox, _("Single image"));
	gtk_signal_connect(GTK_OBJECT(pd->one_image_toggle), "clicked",
			   GTK_SIGNAL_FUNC(spectrum_edit_page_one_image_cb), pd);
	pd->dynamic_toggle = ui_edit_toggle_new(vbox, _("Dynamic size"));

	pd->image_entry = ui_edit_path_entry_new(vbox, _("Image:"), "SLIK_spectrum_image");
	pd->pixmap = ui_edit_pixmap_new(vbox);
	ui_edit_path_entry_connect_pixmap(pd->image_entry, pd->pixmap);

	spectrum = gtk_button_new_with_label(_("Add"));
	gtk_box_pack_start(GTK_BOX(vbox), spectrum, FALSE, FALSE, 0);
	gtk_signal_connect(GTK_OBJECT(spectrum), "clicked",
			   GTK_SIGNAL_FUNC(spectrum_edit_page_add_cb), pd);
	gtk_widget_show(spectrum);

	pd->clist = ui_edit_clist_new(hbox, titles, 4);
	gtk_signal_connect(GTK_OBJECT(pd->clist), "select_row",
			   GTK_SIGNAL_FUNC(spectrum_edit_page_clist_cb), pd);

	gtk_widget_show(hbox);

	return hbox;
}

static void spectrum_edit_page_add(GtkWidget *widget, gpointer data)
{
	SpectrumListData *id = data;
	SpectrumPage *pd;
	gchar *buf[5];

	pd = gtk_object_get_data(GTK_OBJECT(widget), "page");

	buf[0] = "";
	buf[1] = id->key;
	buf[2] = ui_edit_int_to_text(id->segments);
	buf[3] = ui_edit_int_to_text(id->bands);
	buf[4] = NULL;

	ui_edit_clist_append(pd->clist, buf, id->image, id);

	g_free(buf[2]);
	g_free(buf[3]);
}

void spectrum_type_init_edit(WidgetObjectData *od)
{
	od->func_get_pixbuf = spectrum_get_pixbuf;

	od->func_edit_write = spectrum_edit_write;

	od->func_edit_read = spectrum_edit_read;
	od->func_edit_free = spectrum_edit_free;

	od->func_edit_props = spectrum_edit_props;

	od->func_edit_page_new = spectrum_edit_page_new;
	od->func_edit_page_add = spectrum_edit_page_add;
}

