/* 
 * Sound Monitor applet for the GNOME panel
 * (C) 2002 John Ellis
 *
 * Author: John Ellis
 *
 * This software is released under the GNU General Public License (GNU GPL).
 * Please read the included file COPYING for more information.
 * This software comes with no warranty of any kind, use at your own risk!
 */

#include "sound-monitor.h"
#include "update.h"
#include "esdcalls.h"
#include "fft.h"
#include "ui2_item.h"
#include "ui2_meter.h"
#include "ui2_main.h"
#include "ui2_scope.h"
#include "ui2_spectrum.h"

#include <math.h>


static gint display_item_status_cb(ItemData *item, const gchar *key, gpointer data)
{
	AppData *ad = data;

	return (gint)ad->esd_status;
}

static gint display_item_misc_cb(ItemData *item, const gchar *key, gpointer data)
{
	return 0;
}

static gfloat display_meter_misc_cb(MeterData *meter, const gchar *key, gpointer data)
{
	return 0.0;
}

static void display_meter_misc_mode_cb(MeterData *meter, const gchar *key, gpointer data)
{
	AppData *ad = data;
	MeterPeakType mode;

	switch (ad->peak_mode)
		{
		case PeakMode_Active:
			mode = METER_PEAK_ACTIVE;
			break;
		case PeakMode_Smooth:
			mode = METER_PEAK_SMOOTH;
			break;
		default:
			mode = METER_PEAK_OFF;
			break;
		}

	meter_set_peak_mode(meter, mode);
	meter_set_peak_fall(meter, (gfloat)ad->falloff_speed / 10.0);
}

static gint display_spectrum_misc_cb(SpectrumData *spectrum, const gchar *key, gpointer data)
{
	return spectrum_widget_reset(spectrum);
}

static void display_spectrum_misc_mode_cb(SpectrumData *spectrum, const gchar *key, gpointer data)
{
	AppData *ad = data;
	SpectrumPeakType mode;

	switch (ad->peak_mode)
		{
		case PeakMode_Active:
			mode = SPECTRUM_PEAK_ACTIVE;
			break;
		case PeakMode_Smooth:
			mode = SPECTRUM_PEAK_SMOOTH;
			break;
		default:
			mode = SPECTRUM_PEAK_OFF;
			break;
		}

	spectrum_set_peak_mode(spectrum, mode);
	spectrum_set_peak_fall(spectrum, (gfloat)ad->falloff_speed / 10.0);
}

static gint display_scope_misc_cb(ScopeData *scope, const gchar *key, gpointer data)
{
	return scope_widget_reset(scope);
}

static void display_scope_misc_mode_cb(ScopeData *scope, const gchar *key, gpointer data)
{
	AppData *ad = data;

	scope_set_mode(scope, ad->draw_scope_as_segments ? SCOPE_RENDER_LINES : SCOPE_RENDER_POINTS);
	scope_set_scale(scope, ad->scope_scale);
}

void display_register_widgets(AppData *ad)
{

	meter_type_init();
	scope_type_init();
	spectrum_type_init();

	item_register_key("status", ad->ui, display_item_status_cb, ad);

	item_register_key("main", ad->ui, display_item_misc_cb, ad);
	item_register_key("left", ad->ui, display_item_misc_cb, ad);
	item_register_key("right", ad->ui, display_item_misc_cb, ad);

	meter_register_key("main", ad->ui, display_meter_misc_cb, ad,
			   display_meter_misc_mode_cb, ad);
	meter_register_key("left", ad->ui, display_meter_misc_cb, ad,
			   display_meter_misc_mode_cb, ad);
	meter_register_key("right", ad->ui, display_meter_misc_cb, ad,
			   display_meter_misc_mode_cb, ad);

	spectrum_register_key("main", ad->ui, display_spectrum_misc_cb, ad,
			      display_spectrum_misc_mode_cb, ad);
	spectrum_register_key("left", ad->ui, display_spectrum_misc_cb, ad,
			      display_spectrum_misc_mode_cb, ad);
	spectrum_register_key("right", ad->ui, display_spectrum_misc_cb, ad,
			      display_spectrum_misc_mode_cb, ad);

	scope_register_key("main", ad->ui, display_scope_misc_cb, ad,
			   display_scope_misc_mode_cb, ad);
}


void update_modes(AppData *ad)
{
	gfloat fall;
	gint mode;

	/* Note that the meter and spectrum modes are identical,
	 * in fact same as ad->peak_mode, but that may change.
	 */
	switch (ad->peak_mode)
		{
		case PeakMode_Active:
			mode = (gint)METER_PEAK_ACTIVE;
			break;
		case PeakMode_Smooth:
			mode = (gint)METER_PEAK_SMOOTH;
			break;
		default:
			mode = (gint)METER_PEAK_OFF;
			break;
		}

	fall = (gfloat)ad->falloff_speed / 10.0;

	meter_set_options("main", ad->ui, (MeterPeakType)mode, fall);
	meter_set_options("left", ad->ui, (MeterPeakType)mode, fall);
	meter_set_options("right", ad->ui, (MeterPeakType)mode, fall);

	spectrum_set_options("main", ad->ui, (SpectrumPeakType)mode, fall);
	spectrum_set_options("left", ad->ui, (SpectrumPeakType)mode, fall);
	spectrum_set_options("right", ad->ui, (SpectrumPeakType)mode, fall);

	scope_set_options("main", ad->ui,
			  ad->draw_scope_as_segments ? SCOPE_RENDER_LINES : SCOPE_RENDER_POINTS,
			  ad->scope_scale);
}

void update_fps_timeout(AppData *ad)
{
	if (ad->update_timeout_id != -1) gtk_timeout_remove(ad->update_timeout_id);
	ad->update_timeout_id = gtk_timeout_add(1000 / ad->refresh_fps, (GtkFunction)update_display, ad);
	update_modes(ad);
}

static void update_status(AppData *ad)
{
	ad->status_check_count++;
	if (ad->status_check_count > ad->refresh_fps) /* 1 second checks should do */
		{
		gint new_status = esd_status(ad->esd_host);
		ad->status_check_count = 0;
		if (ad->esd_status != new_status)
			{
			ad->esd_status = new_status;
			if (ad->esd_status == Status_Error && ad->sound)
				{
				/* problem must have occured, disconnect from esd now! */
				sound_free(ad->sound);
				ad->sound = NULL;
				gnome_error_dialog("Sound Monitor has lost the connection to the Esound daemon,\n this usually means the daemon has exited and can simply be restarted.\n\n Or the daemon has crashed and needs some attention...");
				}
			}
		}

	if (ad->esd_status_menu != ad->esd_status)
		{
		sync_esd_menu_item(ad);
		item_section_set("status", ad->ui, (gint)ad->esd_status);
		}
}

/* result must be array of length 256 (and zero it first) */
static void sound_get_fft(short *source, gint length, gint32 *result, gint channel)
{
	static fft_state *fft = NULL;
	gfloat tmp_out[257];
	sound_sample tmp_src[FFT_BUFFER_SIZE];
	gint p;

	/* The fft functions in fft.[ch] are straight out of xmms 1.2.6 (www.xmms.org)
	 * and the handling in this function is also based on code in xmms.
	 * I never took the time to understand ffts, but this seems to work...
	 */

	if (!fft) fft = fft_init();

	for (p = 0; p < length / 2 / FFT_BUFFER_SIZE; p++)
		{
		gint i, j;
		j = (p * FFT_BUFFER_SIZE * 2 ) + (channel == 0) ? 0 : 1;
		for (i = 0; i < FFT_BUFFER_SIZE; i++)
			{
			tmp_src[i] = (sound_sample)source[j];
			j += 2;
			}

		fft_perform(tmp_src, tmp_out, fft);

		for (i = 0; i < 256; i++)
			{
			gint32 t;
			t = ((gint32)sqrt(tmp_out[i + 1])) >> 8;
			if (t > result[i]) result[i] = t;
			}
		}
}

gint update_display(gpointer data)
{
	AppData *ad = data;

	update_status(ad);

	if (ad->sound)
		{
		gfloat fall;
		gint new_data;
		gint force_all;

		/* If the editor is active, supply data to
		 * all widget types regardless of existance.
		 */
		force_all = (ad->ui->edit != NULL);

		new_data = ad->sound->new_data;

		if (ad->refresh_fps == 0) ad->refresh_fps = 10;	/* sanity */
		fall = 1.0 / (gfloat)ad->refresh_fps;

		if (new_data)
			{
			ad->no_data_count = 0;
			ad->scope_flat = FALSE;
			}
		else
			{
			if ((float)ad->no_data_count / (ad->refresh_fps ? 1 : ad->refresh_fps) < 0.333 )
				{
				/* no new data in .3 sec results in zeroing */
				ad->no_data_count ++;
				}
			else if (!ad->scope_flat)
				{
				ad->scope_flat = TRUE;
				if (ad->has_scope || force_all) scope_value_set("main", ad->ui, NULL, 0);
				}
			}

		if (ad->has_meter || force_all)
			{
			gint l, r;
			gfloat fl, fr, fm;

			l = r = 0;

			if (new_data)
				{
				sound_get_volume(ad->sound, &l, &r);
				}

			fl = (gfloat)l / 100.0;
			fr = (gfloat)r / 100.0;
			fm = (fl > fr) ? fl : fr;

			item_section_set_by_percent("main", ad->ui, fm);
			item_section_set_by_percent("left", ad->ui, fl);
			item_section_set_by_percent("right", ad->ui, fr);

			meter_value_set("main", ad->ui, fm, fall);
			meter_value_set("left", ad->ui, fl, fall);
			meter_value_set("right", ad->ui, fr, fall);
			}

		if ((ad->has_scope || force_all) && !ad->scope_flat && new_data)
			{
			short *buffer;
			gint length;

			sound_get_buffer(ad->sound, &buffer, &length);
			scope_value_set("main", ad->ui, buffer, length);
			}
		if (ad->has_spectrum || force_all)
			{
			if (new_data)
				{
				short *source;
				gint length;
				gint32 left[SPECTRUM_BUFFER_SIZE];
				gint32 right[SPECTRUM_BUFFER_SIZE];
				gint i;

				sound_get_buffer(ad->sound, &source, &length);

				/* time to burn those cycles :) */

				/* clear buffer */
				memset(left, 0, SPECTRUM_BUFFER_SIZE * sizeof(gint32));
				memset(right, 0, SPECTRUM_BUFFER_SIZE * sizeof(gint32));

				sound_get_fft(source, length, left, 0);
				sound_get_fft(source, length, right, 1);

				spectrum_value_set("left", ad->ui, left, fall);
				spectrum_value_set("right", ad->ui, right, fall);

				/* main channel needs a combine (and more cpu cycles :/) */
				for (i = 0; i < SPECTRUM_BUFFER_SIZE; i++)
					{
					if (right[i] > left[i]) left[i] = right[i];
					}

				spectrum_value_set("main", ad->ui, left, fall);
				}
			else
				{
				spectrum_value_set("main", ad->ui, NULL, fall);
				spectrum_value_set("left", ad->ui, NULL, fall);
				spectrum_value_set("right", ad->ui, NULL, fall);
				}
			}

		ad->sound->new_data = FALSE;
		}

	return TRUE;
}

